// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// snap_rx.cc an implementation of the snap protocol for Linux.
//
// Carl Walker 02-06-99
//
// Released under GPL
//
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//
//
// Overview of how the class works:
//
// Get a byte at a time from the code that handles the serial port.
// (not in this class)
//
// See if it's the SYNC byte, if so start putting the bytes in the
// holding buffer.
//
// Read two more bytes.
//
// Now decode DAB size, SAB size, PFB size, ACK state, CMD mode,
// Error Correction mode, and number of data bytes.
//
// Write the first two bytes into the buffer THROUGH THE CORRECT
// error correction class.
//
// The total number of bytes to read and buffer will be n + data, where
// 0 <= n <= 9 + data_len
// n is comprised of DAB size, SAB size, PFB size, and Error Correction size.
//
// After reading and storing n + data_len bytes, read the error value
// from the error class for the packet. It should be zero. If not, it is
// a bad packet, and should be dropped.
//
// There is a lot more detail that needs to be documented.....
//
// Note that this needs much work to be robust when dealing with
// corrupt packets. The protocol may need a little help here as well,
// but I will not know until moving a lot of data with noise injected
// into the system.
//
// I also need to think about double buffering packets, so that one that
// has been received but not yet read by the owner of this class will not
// cause this class to drop the next packet. Should not be too big a deal.
//

#include "snap_rx.h"

snap_rx::snap_rx( int my_address, int debug ) {

  init_rx();
  auto_dump = 1; // dump packets that are not for me by default
  version = 0.1;
  my_node_address = my_address;
  debug_level = debug;

  // The stats keeping stuff
  stat_bytes = 0;
  stat_good_packets = 0;
  stat_bad_packets = 0;
  stat_payload_bytes = 0;
  stat_dropped_packets = 0;
  stat_broadcast_packets = 0;
  //
  if ( debug_level >= 2 ) {
    cout << "snap_rx::snap_rx:  snap_rx class constructed as object " 
	 << hex << this << " with node address " << dec 
	 << my_node_address << endl;
  }

}

snap_rx::~snap_rx() {
  if ( debug_level >= 2 ) {
    cout << "snap_rx::~snap_rx: snap_rx class destructor called on object " 
	 << hex << this << dec << endl;
  }
}

void snap_rx::init_rx( void ) {
  snap_rx::sync_detected = 0;
  snap_rx::sab_data = 0;
  snap_rx::dab_data = 0;
  snap_rx::pfb_data = 0;
  snap_rx::dab_size = 0;
  snap_rx::sab_size = 0;
  snap_rx::pfb_size = 0;
  snap_rx::ack_state = 0;
  snap_rx::cmd_mode = 0;
  snap_rx::data_len = 0;
  snap_rx::packet_ready = 0;
  snap_rx::total_header_length = 0;
  snap_rx::error_length = 0;
  snap_rx::rx_buffer_offset = 0;
  snap_rx::byte_in_count = 0;
  snap_rx::calc_header = 0;
  snap_rx::err_len = 0;
  snap_rx::err_state = 0;
  snap_rx::total_packet_length = 0;
  snap_rx::starting_offset = 0;
  snap_rx::reported_buffer_len = 0;
  snap_rx::my_packet = 0;
  memset( (void *)&rx_buffer, 0, RX_BUFFER_SIZE );
}

//Accessors
unsigned char snap_rx::packet_ready_query( void ) {
  return packet_ready;
}

int snap_rx::packet_out_len( void ) {
  // Return the length of the packet if it is ready
  if ( packet_ready_query() ) {
    return reported_buffer_len;
  }
  // Return -1 for the length if the packet is not ready to read
  else {
    return ( -1 );
  }
}

//Modifiers
void snap_rx::set_auto_dump( int auto_dump_in ) {
  auto_dump = auto_dump_in;
}

void snap_rx::byte_in ( unsigned char byte_in ) {
  // bump the byte counter for overall stats
  stat_bytes+= 1;

  // Get a byte and see what to do...
  if ( ( sync_detected == 0 ) &
       ( byte_in != SYNC ) ) {
    return; //Not a byte to deal with, just noise or other crap
  }

  if ( ( sync_detected == 0 ) &
       ( byte_in == SYNC ) ) {
    // Here is our start of packet SYNC byte.
    sync_detected = 1;
    return;
  }

  // We are past SYNC, or rejected a non SYNC byte by this point, so it is
  // a real packet byte from here on...

  // bump the byte in counter
  byte_in_count++;

  // Debug message
  if ( debug_level > 5 ) {
    cout << "snap_rx::byte_in: Byte 0x" << hex << (unsigned int)byte_in
	 << " Total " << dec << byte_in_count << " Bytes " << endl;
  }
  
  if ( ( sync_detected ) & ( rx_buffer_offset < 2 ) ) {
    rx_buffer[ rx_buffer_offset ] = byte_in;
    rx_buffer_offset++;
    return;
  }
  
    // At this point we have enough information in the two
    // header bytes to calculate the size of the packet.
    if ( calc_header == 0 ) {
      // Figure out what the total length of all the stuff is going
      // to be
      calc_mode_and_size();
      // do the error stuff for the first two bytes now that
      // we know what the error correction mode is
      do_error( rx_buffer[ 0 ] );
      do_error( rx_buffer[ 1 ] );
    }

    // at this point, we know enough about the packet to process the inbound
    // byte stream
    do_error( byte_in );
    rx_buffer[ rx_buffer_offset ] = byte_in;
    // This will keep us from overflowing the RX buffer
    // if things go bad...
    if ( rx_buffer_offset < RX_BUFFER_SIZE )
      rx_buffer_offset++;

    // Now we check and see if that was the last byte...
    if ( rx_buffer_offset <  total_packet_length ) {
      // not done yet
      return;
    }
    else {
      // We just got the last byte!
      // so we check on the error status of the packet
      // count as a good packet, EVEN IF NOT.
      stat_good_packets += 1;
      // confirm_error() will drop the good packet count back down
      // if this packet is corrupt.
      confirm_error();
      // extract the sab, dab, and pfb data bytes and store them
      extract_header_data();
      // see if this packet was for me
      if ( dab_data == my_node_address ) {
	my_packet = 1;
      }
      // or if was a broadcast packet
      // NOTE THAT PACKETS WITH A ZERO LENGTH DAB FIELD ARE BROADCAST
      // BY THEIR VERY NATURE!
      else if ( ( dab_data == BROADCAST ) | ( dab_size == ADDR_0_BIT ) ) {
	my_packet = 1;
	broadcast = 1;
	stat_broadcast_packets += 1;
      }
      // Now we must calculate the start offset of the data, and
      // the end point of the buffer without the error correction byte(s)
      starting_offset = total_header_length;
      reported_buffer_len = total_packet_length - 
	total_header_length - err_len;
      // and update the payload stats
      stat_payload_bytes += reported_buffer_len;

      // show ready
      packet_ready = 1;
      
      if ( debug_level >= 3 ) {
	cout << "snap_rx:: PACKET READY -- Header Length: " 
	     << total_header_length << " Packet Length: "
	     << total_packet_length << " Error Length: "
	     << err_len << endl;
	display_raw_packet(); // show the world what we got
      }
    }
    // automaticly clear the packet if it is not for me
    // (if that's enabled)
    if ( auto_dump && ( my_packet == 0 ) ) {
      // bump the stats for dropped packets
      stat_dropped_packets+= 1;
      // Clear everything out from the receiver
      init_rx();
      if ( debug_level >= 3 ) {
	cout << "snap_rx:: Packet not addressed to me. Dumping." << endl;
      }
    }
}

unsigned char * snap_rx::packet_out( void ) {
  return (unsigned char *)&rx_buffer[ starting_offset ];
}

void snap_rx::calc_mode_and_size( void ) {
  // HDB2 is in rx_buffer[ 0 ] and HDB1 is in [ 1 ].....
  // Figure out how big the transfer will be!
  unsigned char hdb2 = rx_buffer[ 0 ];
  unsigned char hdb1 = rx_buffer[ 1 ];
  if ( debug_level >= 5 ) {
    cout << "snap_rx::calc_mode_and_size: HDB2: " << hex << (unsigned int)hdb2
	 << " HDB1: " << (unsigned int)hdb1 << dec << endl;
  }
  dab_size =  ( ( hdb2 & 0xc0 ) >> 6 );
  sab_size =  ( ( hdb2 & 0x30 ) >> 4 );
  pfb_size =  ( ( hdb2 & 0x0c ) >> 2 );
  ack_state = ( hdb2 & 0x03 );
  cmd_mode =  ( ( hdb1 & 0x80 ) >> 7 );
  err_mode =  ( ( hdb1 & 0x70 ) >> 4 );
  data_len =  calc_ndb( ( hdb1 & 0x0f ) );
  if ( data_len < 0 ) {
    cout << "FATAL ERROR: DATA LENGTH < 0" << endl;
    // FOR SOME REASON, SLOWLARIS DOES NOT KNOW 
    //ABOUT THE PROTOTYPE FOR exit(int);
    //exit(0);
  }

  // now we can figure out how big this sucker is going to be
  total_header_length = dab_size + sab_size + pfb_size + 2;
  // +2 comes from the fact that we always have HDB2 and HDB1
  // in the protocol! We have to count them in the total header length

  //now figure out how many bytes are in the error correction
  switch ( err_mode ) {
  case NO_ERR_MODE:
    break;
  case MULT_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case BYTE_CKSUM_ERR_MODE:
    err_len = 1;
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case WORD_CRC_ERR_MODE:
    err_len = 2;
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case FEC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_err_mode );
    break;    
  }
  // Figure out how much data there is in the incoming packet with
  // headers and data and error correction
  total_packet_length = total_header_length + data_len + err_len;
  // And now we know!
  calc_header = 1; // flag that we have already done this
}

int snap_rx::calc_ndb( unsigned char packet_ndb ) {
  // Figure out the size of the payload
  if ( packet_ndb <= 8 ) {
    return packet_ndb;
  }
  switch ( packet_ndb ) {
  case BYTE_16_LEN:
    return 16;
  case BYTE_32_LEN:
    return 32;
  case BYTE_64_LEN:
    return 64;
  case BYTE_128_LEN:
    return 128;
  case BYTE_256_LEN:
    return 256;
  case BYTE_512_LEN:
    return 512;
  }
  // return -1 if we did not match a size in the switch statement
  // ( which is impossible, but this will keep g++ from complaining )
  return -1;
}

void snap_rx::do_error( unsigned char byte_in ) {
  switch ( err_mode ) {
  case NO_ERR_MODE:
    break;
  case MULT_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case BYTE_CKSUM_ERR_MODE:
    checksum_8->calc_checksum8( byte_in );;
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case WORD_CRC_ERR_MODE:
    crc_16->calc_crc16( byte_in );
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case FEC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_err_mode );
    break;    
  }
}

void snap_rx::confirm_error( void ) {
  int error_return = 0;
  switch ( err_mode ) {
  case NO_ERR_MODE:
    // If there is no error correction, the packet is always 'good'
    break;
  case MULT_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case BYTE_CKSUM_ERR_MODE:
    error_return = checksum_8->return_checksum8();
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case WORD_CRC_ERR_MODE:
    error_return = crc_16->return_crc16();
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case FEC_ERR_MODE:
    report_function_error( e_err_mode );
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_err_mode );
    break;    
  }
  if ( error_return ) {
    // bump the bad packet counter
    stat_bad_packets += 1;
    // and the good packet counter
    stat_good_packets -= 1;

    // Must have been an error, as the final calculation did not return zero
    if ( debug_level >= 3 ) {
      cout << "snap_rx::confirm_error: BAD PACKET! " << hex
	   << error_return << dec << " return value " << endl;
    }
    err_state = 1;
  }
}

void snap_rx::extract_header_data( void ) {
  int i;
  // Get the sab, dab, and pfb information from the header and
  // store them so that they may be accessed
  //
  // ONLY get the field if it has a size greater than zero!
  if ( dab_size > 0 ) {
    for ( i = 1; i <= dab_size; i++ ) {
      dab_data |= rx_buffer[ (1 + i) ]; // The 1 comes from the fact that there
      // are already two header [0-1] bytes hdb2 and hdb1 at the 
      // start of the buffer
      dab_data = dab_data << 8; // post shift the data for the next byte
    }
    dab_data = dab_data >> 8; // shift it back. Yes this is inefficient for
    // single byte data. Got any better ideas?
  }

  if ( sab_size > 0 ) {
    for ( i = 1; i <= sab_size; i++ ) {
      sab_data |= rx_buffer[ (1 + i + dab_size) ];
      sab_data = sab_data << 8; // post shift the data for the next byte
    }
    sab_data = sab_data >> 8; // shift it back. 
  }

  if ( pfb_size > 0 ) {
    for ( i = 1; i <= pfb_size; i++ ) {
      pfb_data |= rx_buffer[ (1 + i + dab_size + sab_size) ];
      pfb_data = pfb_data << 8; // post shift the data for the next byte
    }
    pfb_data = pfb_data >> 8; // shift it back. 
  }

}

int snap_rx::get_dab_data( void ) {
  return dab_data;
}

int snap_rx::get_sab_data( void ) {
  return sab_data;
}

int snap_rx::get_pfb_data( void ) {
  return pfb_data;
}

void snap_rx::show_stats( void ) {
  // report on the stats that the snap_rx class has gathered while running
  cout << endl;
  cout << "Stats for snap_rx Object " << hex << this << dec << endl << endl;
  cout << "Total       Bytes:        " << stat_bytes << endl;
  cout << "Payload     Bytes:        " << stat_payload_bytes << endl;
  cout << "Protocol Overhead:        " <<
    ( ( (float)( stat_bytes - stat_payload_bytes ) / (float)stat_bytes  ) 
      * 100 ) 
       << " %"<< endl;
  cout << "Good      Packets:        " << stat_good_packets << endl;
  cout << "Bad       Packets:        " << stat_bad_packets << endl;
  cout << "Broadcast Packets:        " << stat_broadcast_packets << endl;
  cout << "Dropped   Packets:        " << stat_dropped_packets << endl;
  cout << endl;
}

void snap_rx::display_raw_packet( void ) {

  int i = 0;
  // print out the raw data stream for the packet that we just got
    for ( ; i < total_packet_length; i++ ) {
      cout << setfill( '0' ) << setw( 2 );
      cout << hex << (unsigned int)rx_buffer[ i ] << " ";
    }
    cout << endl;

}
