// ******************************************************************************
// An implementation of the snap protocol in C++
//
// Carl Walker
// ICMP Networks
// 64 West Hollis Road
// Hollis, NH 03049 USA
// snap@icmp.com
//
// Released under GPL
//
// 02-04-99
// ******************************************************************************
// Functions implemented:
//
// get_version(void)         Returns the current version (float)*
// set_debug_level(int)      sets a debugging level for message generation*
// get_debug_level(void)     Returns the current debugging level as an int*
// set_sab_len(int)          Sets how many bytes are in the Source Address*
// set_dab_len(int)          Sets how many bytes are in the Destination Address*
// set_pfb_len(int)          Sets how many bytes are in the Protocol specific flags*
// set_pfb_data(int)         Sets the pfb data, masked always to lower 24 bits*
//
// set_ack_type(int)         Set up the required type of ACK in the header*
// set_data_length(int)      Sets the header field for the transmit packet*
// set_node_address(int)     Sets the node address of this master/slave*
// get_node_address(void)    Returns int current node address*
// comp_tx_packet(void)      Compiles a packet from the currently set
//                           data frame, dest address, and other parameters
//                           and returns a pointer to a compiled packet*
// get_packet_len(void)      returns the total length of the compiled packet*
// tx_packet_done(void)      Tells the class that the pointer returned from
//                           the last build request has been sent, and the buffer
//                           is now free*
// set_dab_data(int)         Sets the address to send the packet to*
// get_dab_data(void)        Returns current dest address int*
// set_edm(int)              Sets the error correction mode*
// get_edm(void)             Returns as an int the current error mode*
// tx_data(char *, int len)  Pass a buffer of chars to send to the remote node
//                           and the length of the data.
//                           returns the length of the buffer if all is well,
//                           -1 on error. NULL, NULL can be used to assemble
//                           a packet with a zero data length*
// ******************************************************************************
//

#include "snap_defines.h"
#include "snap_tx.h"

// Constructor
snap_tx::snap_tx( int local_node_address, int debug ) {
  version = 0.1;
  tx_edm_mode = 0;           //default to error mode off
  sab_data = local_node_address;
  debug_level = debug;
  tx_packet_buffer_busy = 0; // not busy
  sab_size = dab_size = pfb_size = 0;
  cmd = 0;
  //set up the pointers and the data buffers
  tx_data_ptr   = (char *)&tx_data_buffer;
  tx_packet_ptr = (char *)&tx_packet_out;
  memset( (void *)tx_data_ptr,   0, TX_BUFFER_LEN );
  memset( (void *)tx_packet_out, 0, TX_BUFFER_LEN );

  if ( debug_level >= 2 ) {
    cout << "snap_tx::snap_tx:  snap_tx class constructed as object " 
	 << hex << this << " with node address " << dec 
	 << my_node_address << endl;
  }
}

// Destructor
snap_tx::~snap_tx(){
  if ( debug_level >= 2 ) {
    cout << "snap_tx::~snap_tx: snap_tx class destructor called on object " 
	 << hex << this << dec << endl;
  }
}

// Modifiers
void snap_tx::set_edm( int error_mode_in ) {
  err_mode = ( error_mode_in & ERROR_MASK );
}

void snap_tx::set_dab_data( int dest_addr_in ) {
  dab_data = dest_addr_in;
}

void snap_tx::tx_packet_done(void) {
  tx_packet_buffer_busy = 0;
  tx_packet_len = 0;

  raw_tx_byte_count = 1; 
  // It is '1' because we start counting the SYNC byte
  // that never goes through one of the error / CRC functions that are normally used
  // to count the tx bytes in the packet. We only really use raw_tx_byte_count for
  // packets that have NO error correction mode!
}

void snap_tx::set_sab_len(int length_in) {
  sab_size = ( length_in & TWO_BIT_MASK );
}

void snap_tx::set_dab_len(int length_in) {
  dab_size = ( length_in & TWO_BIT_MASK );
}

void snap_tx::set_pfb_len(int length_in) {
  pfb_size = ( length_in & TWO_BIT_MASK );
}

void snap_tx::set_pfb_data( int pfb_data_in ) {
  pfb_data = ( pfb_data_in & THREE_BYTE_MASK );
}

void snap_tx::set_cmd_mode(int command_in) {
  cmd = ( command_in & 0x1 );
}

void snap_tx::set_ack_type( int ack_in ) {
  ack_type = ( ack_in & TWO_BIT_MASK );
}

// Private modifiers
void snap_tx::set_data_len( int length ) {
  tx_data_length = ( length & NIBBLE_MASK );
}

// Accessors
int snap_tx::get_edm( void ) {
  return err_mode;
}

int snap_tx::get_dab_data( void ) {
  return dab_data;
}

// 'real' functions for the snap class
char* snap_tx::comp_packet( void ) {
  int i = 0;
  int temp;
  // OK, build a transmit packet from the infromation and the data that
  // we have.....
  if ( tx_packet_buffer_busy ) {
    //can't use the tx buffer yet, it is not empty

    if ( debug_level >= 2 ) {
      cout << "snap_tx::comp_packet: Called while buffer is not empty" << endl;
    }

    return NULL; //so we return a null pointer
  }

  // Clear any and all transmit CRC / Checksum calculators
  crc_16->init_crc16();
  checksum_8->init_checksum8();

  // Looks good, here we go...
  add_sync();           // start with the sync byte
  generate_header(); // add the header definition bytes
                        // and the dest and source address bytes
  set_address_bytes( sab_size, dab_size ); 

  // *********************************************
  // ***** Now we have to deal with the data *****
  // *********************************************

  for ( i = 0; i < tx_data_length_buffer; i++ ) {
    // copy the buffer through the error check / generation functions
    temp = tx_data_buffer[ i ];
    write_packet_byte( temp );
  }
  // Done with the buffer copy

  //Now we deal with the error bytes
  write_packet_error_bytes();
  // We will now get the total length of the packet we just sent.
  // Look in the function write_packet_error_bytes() to see how it works.
  int total_bytes = 0;
  switch ( err_mode ) {

  case NO_ERR_MODE:
    total_bytes = raw_tx_byte_count;
    break;

  case BYTE_CKSUM_ERR_MODE:
    total_bytes = get_packet_len();
    break;

  case WORD_CRC_ERR_MODE:
    total_bytes = get_packet_len();
    break;

  default:
    cout << "snap_tx::comp_packet: Default case reached while calculating length " 
	 << endl;
  }
  if ( debug_level >= 2 ) {
    cout << "snap_tx::comp_packet: Generated a " << dec << total_bytes <<
      " byte Transmit packet" << endl;
  }
  return (char *)&tx_packet_out;
}

void snap_tx::generate_header( void ) {
  int temp = 0;
  // build and store HDB2 and HDB1 from the stuff that we have
  temp = ( dab_size << 6 ) | ( sab_size << 4 ) | 
    ( pfb_size << 2 ) | ( ack_type );
  // store HDB2
  write_packet_byte( temp );
  // build HDB1
  temp = 0;
  // Need to switch on packet payload sizes greater than 8 bytes...
  // Size of packet > 8 bytes, and padding must be enforced by the tx_data(...)
  // function
  if ( tx_data_length_buffer <= 8 ) {
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( tx_data_length_buffer ); 
  }
  else {
    switch ( tx_data_length_buffer ) {
    case 16:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_16_LEN ); 
      break;
    case 32:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_32_LEN ); 
      break;
    case 64:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_64_LEN ); 
      break;
    case 128:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_128_LEN ); 
      break;
    case 256:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_256_LEN ); 
      break;
    case 512:
    temp = ( cmd << 7 ) | ( err_mode << 4 ) |
      ( BYTE_512_LEN ); 
      break;
    default:
      cout << "snap_tx::generate_header: Default case reached. Packet is mangled"
	   << endl;
      break;
    }
  }
  //tx_data_length_buffer is how much payload data
  //there is in the packet, not the total byte count of the packet. It comes from
  //the tx_data(...) function that gets the data payload from the outside world
  //well, outside of this class, anyway...
  // Now store HDB1
  write_packet_byte( temp );
}

void snap_tx::set_address_bytes( int source_byte_len, int dest_byte_len ) {
  //deal with the source bytes
  address_byte_out( dest_byte_len, dab_data );
  //deal with the destination bytes
  address_byte_out( source_byte_len, sab_data );
  //add the pfb byte(s), if needed
  if ( pfb_size > 0 ) {
    address_byte_out( pfb_size, pfb_data );
  }
}

void snap_tx::add_sync( void ) {
  tx_packet_out_offset = 0; //zero the offset for the packet that we are building
  tx_packet_out[ tx_packet_out_offset ] = SYNC;  //add the sync byte

  if ( debug_level >= 3 ) {
    // print out a bunch of stuff about the packet as we build it
    cout << "snap_tx::add_sync:             Offset: " << tx_packet_out_offset <<
      " Byte: 0x" << hex << (int)tx_packet_out[ tx_packet_out_offset ] 
	 << dec <<
      " Raw Count: " << raw_tx_byte_count <<
      " --- " << endl;
  }

  tx_packet_out_offset++; //and bump the pointer
}

void snap_tx::address_byte_out( int len, int addr ) {
  //take the length of the address and the address itself and stuff them
  switch (len) {

  case 0: //a zero byte length? OK.....
    break;

  case 1: // 8 bit address
    write_packet_byte( addr & ONE_BYTE_MASK );
    break;

  case 2: // 16 bit address
    write_packet_byte( ( addr & ( ONE_BYTE_MASK << 8 ) ) >> 8 );
    write_packet_byte( addr & ONE_BYTE_MASK );
    break;

  case 3: // 24 bit address
    write_packet_byte( ( addr & ( ONE_BYTE_MASK << 16 ) ) >> 16 );
    write_packet_byte( ( addr & ( ONE_BYTE_MASK << 8 ) ) >> 8 );
    write_packet_byte( addr & ONE_BYTE_MASK );
    break;

  default:
    cout << "snap_tx::address_byte_out: Illegal Address Length! 0x" 
	 << hex << len << dec << endl;
    cout << "snap_tx::address_byte_out: No address stuffed." << endl;
  }
}

void snap_tx::write_packet_byte( unsigned char byte_to_write ) {
  tx_packet_out[ tx_packet_out_offset ] = byte_to_write;
  calc_error( byte_to_write );

  if ( debug_level >= 3 ) {
    // print out a bunch of stuff about the packet as we build it
    cout << "snap_tx::write_packet_byte: Offset: " 
	 << tx_packet_out_offset << " Byte: 0x" << hex 
	 << (int)byte_to_write << dec << " Raw Count: " 
	 << raw_tx_byte_count << " --- " << endl;
  }
  tx_packet_out_offset++; // bump the counters
  raw_tx_byte_count++;
}

void snap_tx::calc_error( unsigned char byte_in ) {
  //Figure out what type of error checking we are using, and call the
  //correct error calculator with this byte tht we just got from the
  //write_packet_byte() function
  switch ( err_mode ) {
  case NO_ERR_MODE:
    break; //do nothing without error checking
  case MULT_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case BYTE_CKSUM_ERR_MODE:
    checksum_8->calc_checksum8( byte_in );
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case WORD_CRC_ERR_MODE:
    crc_16->calc_crc16( byte_in );
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case FEC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  default:
    cout << "snap_tx::calc_error: Default case reached. Something is broken" <<
      endl;
  }
}

void snap_tx::write_packet_error_bytes( void ) {
  int temp = 0;
  int temp_byte_count = 0;
  //Figure out what type of error checking we are using, and call the
  //correct error calculator and get the error byte(s) from it. Write
  //out the proper nymber of bytes to the tx_packet_buffer
  switch ( err_mode ) {
  case NO_ERR_MODE:
    break; //do nothing without error checking
  case MULT_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case BYTE_CKSUM_ERR_MODE:
    temp = checksum_8->return_checksum8();
    temp_byte_count = checksum_8->return_byte_count();
    write_packet_byte( temp );
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case WORD_CRC_ERR_MODE:
    // get the CRC word
    temp = crc_16->return_crc16();
    temp_byte_count = crc_16->return_byte_count();
    //write the hi byte
    write_packet_byte( ( temp >> 8 ) );
    //and the low byte
    write_packet_byte( ( temp & ONE_BYTE_MASK ) );
    // The 'real' number of bytes in the in the packet and header is in
    // temp_byte_count now, and we wrote the CRC bytes through the calculation
    // function. It seems funny ( and it is ), as the calculation now includes
    // the CRC data itself. Therefor, the CRC value is no longer any good. However,
    // the byte length function will have the 'real' byte count of the packet
    // including the error bytes, but excluding the SYNC byte.
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case FEC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  default:
    cout << "snap_tx::write_packet_error_bytes: " <<
      "Default case reached. Something is broken" <<
      endl;
  }
}

int snap_tx::get_packet_len( void ) {
  switch ( err_mode ) {
  case NO_ERR_MODE:
    tx_packet_len = raw_tx_byte_count;
    break;
  case MULT_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case BYTE_CKSUM_ERR_MODE:
    //REMEMBER TO ADD IN THE SYNC BYTE TO THE COUNT
    tx_packet_len = ( ( checksum_8->return_byte_count() ) + 1 );
    break;
  case BYTE_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case WORD_CRC_ERR_MODE:
    // REMEMBER TO ADD IN THE SYNC BYTE HERE! (Thats where the '+ 1' comes from)
    tx_packet_len = ( ( crc_16->return_byte_count() ) + 1 );
    break;
  case DWORD_CRC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case FEC_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  case UDEF_ERR_MODE:
    report_function_error( e_general ); // report that we don't support this
    break;
  default:
    cout << "snap_tx::get_packet_len: " <<
      "Default case reached. Something is broken" <<
      endl;
  }
  return tx_packet_len;
}

int snap_tx::tx_data( char * data_in_ptr, int len ) {
  int i;
  
  if ( len > TX_BUFFER_LEN ) {
    // make sure that we don't overflow the buffer with a packet larger than it
    len = ( TX_BUFFER_LEN - 1 );
    cout << "snap_tx::tx_data: Please don't overflow the transmit packet buffer."
	 << " Packet truncated to " << TX_BUFFER_LEN << " bytes." << endl;
  }
  
  // Store the byte length we need. If the payload is larger than 8 bytes, 
  // we will change this value later in the function to accomadate the 
  // corse packet sizing and padding. Read on.

  // store the length of the information in this buffer
  tx_data_length_buffer = len;

  // empty the tramsmit buffer, and fill with zeros so that padding will work on
  // larger than 8 byte packets
  memset( (void *)tx_packet_out, 0x0, TX_BUFFER_LEN );

  // copy the data...
  // only do the copy if we have a valid data pointer and length
  if ( ( data_in_ptr != NULL ) & ( len > 0 ) ) { 
    for ( i = 0; i < len; i++ ) {
      tx_data_buffer[ i ] = ( data_in_ptr[ i ] & ONE_BYTE_MASK );
    }
  }

  // enforce the sizes required by packets > 8 bytes
  if ( ( len > 8 ) & ( len <= 16 ) ) {
    tx_data_length_buffer = 16;
  }
  if ( ( len > 16 ) & ( len <= 32 ) ) {
    tx_data_length_buffer = 32;
  }
  if ( ( len > 32 ) & ( len <= 64 ) ) {
    tx_data_length_buffer = 64;
  }
  if ( ( len > 64 ) & ( len <= 128 ) ) {
    tx_data_length_buffer = 128;
  }
  if ( ( len > 128 ) & ( len <= 256 ) ) {
    tx_data_length_buffer = 256;
  }
  if ( ( len > 256 ) & ( len <= 512 ) ) {
    tx_data_length_buffer = 512;
  }
  if ( len > 512 ) {
    cout << "snap_tx::tx_data: Packet payload length of " 
	 << len << " is too large!"
	 << " Data will be truncated to 512 Bytes" << endl;
    tx_data_length_buffer = 512;
  }
  return (tx_data_length_buffer);
}
